import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
import h5py
from typing import Dict

import logging
logger = logging.getLogger(__name__)

from pegasusio import UnimodalData, MultimodalData



def load_10x_h5_file_v2(h5_in: h5py.Group, ngene: int = None) -> MultimodalData:
    """Load 10x v2 format matrix from hdf5 file

    Parameters
    ----------

    h5_in : h5py.Group
        An instance of h5py.Group class that is connected to a 10x v2 formatted hdf5 file.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file_v2(h5_in)
    """
    data = MultimodalData()
    for genome in h5_in.keys():
        group = h5_in[genome]

        M, N = group["shape"][...]
        mat = csr_matrix(
            (
                group["data"][...],
                group["indices"][...],
                group["indptr"][...],
            ),
            shape=(N, M),
        )

        barcodes = group["barcodes"][...].astype(str)
        ids = group["genes"][...].astype(str)
        names = group["gene_names"][...].astype(str)

        unidata = UnimodalData({"barcodekey": barcodes}, 
        	{"featurekey": names, "featureid": ids}, 
        	{"X": mat}, 
        	metadata = {"experiment_type": "rna", "genome": genome}
        )
        unidata.filter(ngene=ngene)
        unidata.separate_channels()

        data.add_data(genome, unidata)

    return data


def load_10x_h5_file_v3(h5_in: h5py.Group, ngene: int = None) -> MultimodalData:
    """Load 10x v3 format matrix from hdf5 file

    Parameters
    ----------

    h5_in : h5py.Group
        An instance of h5py.Group class that is connected to a 10x v3 formatted hdf5 file.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file_v3(h5_in)
    """
    M, N = h5_in["matrix/shape"][...]
    bigmat = csr_matrix(
        (
            h5_in["matrix/data"][...],
            h5_in["matrix/indices"][...],
            h5_in["matrix/indptr"][...],
        ),
        shape=(N, M),
    )
    barcodes = h5_in["matrix/barcodes"][...].astype(str)
    genomes = h5_in["matrix/features/genome"][...].astype(str)
    ids = h5_in["matrix/features/id"][...].astype(str)
    names = h5_in["matrix/features/name"][...].astype(str)

    data = MultimodalData()
    for genome in np.unique(genomes):
        idx = genomes == genome

        barcode_metadata = {"barcodekey": barcodes}
        feature_metadata = {"featurekey": names[idx], "featureid": ids[idx]}
        mat = bigmat[:, idx].copy()
        unidata = UnimodalData(barcode_metadata, feature_metadata, {"X": mat}, metadata = {"experiment_type": "rna", "genome": genome})
        unidata.filter(ngene=ngene)
        unidata.separate_channels()

        data.add_data(genome, unidata)

    return data


def load_10x_h5_file(input_h5: str, ngene: int = None) -> MultimodalData:
    """Load 10x format matrix (either v2 or v3) from hdf5 file

    Parameters
    ----------

    input_h5 : `str`
        The matrix in 10x v2 or v3 hdf5 format.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file('example_10x.h5')
    """
    data = None
    with h5py.File(input_h5, 'r') as h5_in:
        load_file = load_10x_h5_file_v3 if "matrix" in h5_in.keys() else load_10x_h5_file_v2
        data = load_file(h5_in, ngene)

    return data


def load_pegasus_h5_file(
    input_h5: str, ngene: int = None, select_singlets: bool = False
) -> MultimodalData:
    """Load matrices from pegasus-format hdf5 file (deprecated)

    Parameters
    ----------

    input_h5 : `str`
        pegasus-format hdf5 file.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.
    select_singlets: `bool`, optional (default: False)
        If only load singlets.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_pegasus_h5_file('example.h5sc')
    """
    cite_seq_name = None
    selected_barcodes = None

    data = MultimodalData()
    with h5py.File(input_h5, 'r') as h5_in:
        for genome in h5_in.keys():
            group = h5_in[genome]            

            M, N = group["shape"][...]
            mat = csr_matrix(
                (
                    group["data"][...],
                    group["indices"][...],
                    group["indptr"][...],
                ),
                shape=(N, M),
            )

            barcode_metadata = {}
            sub_group = group["_barcodes"]
            for key in sub_group.keys():
                if key != "barcodekey":
                    continue
                values = sub_group[key][...]
                if values.dtype.kind == "S":
                    values = values.astype(str)
                barcode_metadata[key] = values

            feature_metadata = {}
            sub_group = group["_features"]
            for key in sub_group.keys():
                values = sub_group[key][...]
                if values.dtype.kind == "S":
                    values = values.astype(str)
                if key == "featurename":
                    key = "featurekey"
                elif key == "featurekey":
                    key = "featureid"                    
                feature_metadata[key] = values

            is_citeseq = genome.startswith("CITE_Seq")
            unidata = UnimodalData(barcode_metadata, feature_metadata, {"X": mat}, metadata = {"experiment_type": "citeseq" if is_citeseq else "rna", "genome": genome})

            if is_citeseq:
                cite_seq_name = genome
            else:
                unidata.filter(ngene, select_singlets)
                selected_barcodes = unidata.obs_names

            data.add_data(genome, unidata)

    if (cite_seq_name is not None) and (selected_barcodes is not None):
        unidata = data.get_data(cite_seq_name)
        selected = unidata.obs_names.isin(selected_barcodes)
        unidata.trim(selected)

    return data


def load_loom_file(input_loom: str, genome: str = None, exptype: str = None, ngene: int = None) -> MultimodalData:
    """Load count matrix from a LOOM file.

    Parameters
    ----------

    input_loom : `str`
        The LOOM file, containing the count matrix.
    genome : `str`, optional (default None)
        The genome reference. If None, use "unknown" instead. If not None and input loom contains genome attribute, the attribute will be overwritten.
    exptype: `str`, optional (default None)
        Experiment type. If None, use "rna" instead. If not None and input loom contains experiment_type attribute, the attribute will be overwritten.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes. Only apply to data with exptype == "rna".

    Returns
    -------

    A MultimodalData object containing a (genome, UmimodalData) pair.

    Examples
    --------
    >>> io.load_loom_file('example.loom', genome = 'GRCh38', ngene = 200)
    """
    col_trans = {"CellID": "barcodekey", "obs_names": "barcodekey"}
    row_trans = {"Gene": "featurekey", "var_names": "featurekey", "Accession": "featureid",  "gene_ids": "featureid"}

    data = MultimodalData()

    import loompy
    with loompy.connect(input_loom) as ds:
        barcode_metadata = {}
        barcode_multiarrays = {}
        for key, arr in ds.col_attrs.items():
            key = col_trans.get(key, key)
            if arr.ndim == 1:
                barcode_metadata[key] = arr
            elif arr.ndim > 1:
                barcode_multiarrays[key] = arr
            else:
                raise ValueError("Detected column attribute {} has ndim = {}!".format(key, arr.ndim))

        feature_metadata = {}
        feature_multiarrays = {}
        for key, arr in ds.row_attrs.items():
            key = row_trans.get(key, key)
            if arr.ndim == 1:
                feature_metadata[key] = arr
            elif arr.ndim > 1:
                feature_multiarrays[key] = arr
            else:
                raise ValueError("Detected row attribute {} has ndim = {}!".format(key, arr.ndim))

        matrices = {}
        for key, mat in ds.layers.items():
            key = "X" if key == "" else key
            matrices[key] = mat.sparse().T.tocsr()

        metadata = dict(ds.attrs)
        if genome is not None:
            metadata["genome"] = genome
        elif "genome" not in metadata:
            metadata["genome"] = "unknown"
        if exptype is not None:
            metadata["experiment_type"] = exptype
        elif "experiment_type" not in metadata:
            metadata["experiment_type"] = "rna"
            
        unidata = UnimodalData(barcode_metadata, feature_metadata, matrices, barcode_multiarrays, feature_multiarrays, metadata)
        if metadata["experiment_type"] == "rna":
            unidata.filter(ngene = ngene)

    data.add_data(genome, unidata)

    return data


def write_loom_file(data: MultimodalData, output_file: str) -> None:
    """ Write a MultimodalData to loom file. Will assert data only contain one type of experiment.
    """
    keys = data.list_data()
    if len(keys) > 1:
        raise ValueError("Data contain multiple modalities: {}!".format(",".join(keys)))
    data.select_data(keys[0])
    matrices = data.list_keys()
    assert "X" in matrices
    if len(matrices) == 0:
        raise ValueError("Could not write empty matrix to a loom file!")

    def _process_attrs(key_name: str, attrs: pd.DataFrame, attrs_multi: dict) -> Dict[str, object]:
        res_dict = {key_name: attrs.index.values}
        for key in attrs.columns:
            res_dict[key] = np.array(attrs[key].values)
        for key, value in attrs_multi.items():
            if value.ndim > 1: # value.ndim == 1 refers to np.recarray, which will not be written to a loom file.
                res_dict[key] = value if value.shape[1] > 1 else value[:, 0]
        return res_dict

    row_attrs = _process_attrs("Gene", data.var, data.varm)
    col_attrs = _process_attrs("CellID", data.obs, data.obsm)

    accession_key = "featureid" if "featureid" in row_attrs else ("gene_ids" if "gene_ids" in row_attrs else None)
    if accession_key is not None:
        row_attrs["Accession"] = row_attrs.pop(accession_key)

    layers = {}
    for matkey in matrices:
        layers["" if matkey == "X" else matkey] = data.get_matrix(matkey).T

    file_attrs = {}
    print(type(data.uns))
    for key, value in data.uns.items():
        if isinstance(value, str):
            file_attrs[key] = value

    import loompy
    loompy.create(output_file, layers, row_attrs, col_attrs, file_attrs = file_attrs)

    logger.info("{0} is written.".format(output_file))



import sys
import logging
import warnings


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)

warnings.filterwarnings("ignore", category=FutureWarning, module='anndata')


from .unimodal_data import UnimodalData
from .multimodal_data import MultimodalData
from .readwrite import infer_file_type, read_input, write_output

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, hstack, vstack

from typing import List, Dict, Union, Set
import anndata

from pegasusio import UnimodalData



class MultimodalData:
    def __init__(self, data_dict: Union[Dict[str, UnimodalData], anndata.AnnData] = None, genome: str = None, exptype: str = None):
        if isinstance(data_dict, anndata.AnnData):
            self.from_anndata(data_dict, genome = genome, exptype = exptype)
            return None

        self.data = data_dict if data_dict is not None else dict()
        self._selected = self._unidata = None


    def __repr__(self) -> str:
        repr_str = "MultimodalData object with {} UnimodalData: {}".format(len(self.data), str(list(self.data))[1:-1])
        if self._selected is not None:
            repr_str += "\n    It currently binds to UnimodalData object {}\n\n".format(self._selected)
            repr_str += self._unidata.__repr__()
        else:
            repr_str += "\n    It currently binds to no UnimodalData object"

        return repr_str


    @property
    def obs(self) -> Union[pd.DataFrame, None]:
        return self._unidata.obs if self._unidata is not None else None

    @obs.setter
    def obs(self, obs: pd.DataFrame):
        assert self._unidata is not None
        self._unidata.obs = obs

    @property
    def obs_names(self) -> Union[pd.Index, None]:
        return self._unidata.obs_names if self._unidata is not None else None

    @obs_names.setter
    def obs_names(self, obs_names: pd.Index):
        assert self._unidata is not None
        self._unidata.obs_names = obs_names

    @property
    def var(self) -> Union[pd.DataFrame, None]:
        return self._unidata.var if self._unidata is not None else None

    @var.setter
    def var(self, var: pd.DataFrame):
        assert self._unidata is not None
        self._unidata.var = var

    @property
    def var_names(self) -> Union[pd.Index, None]:
        return self._unidata.var_names if self._unidata is not None else None

    @var_names.setter
    def var_names(self, var_names: pd.Index):
        assert self._unidata is not None
        self._unidata.var_names = var_names

    @property
    def X(self) -> Union[csr_matrix, None]:
        return self._unidata.X if self._unidata is not None else None

    @X.setter
    def X(self, X: csr_matrix):
        assert self._unidata is not None
        self._unidata.X = X

    @property
    def obsm(self) -> Union[Dict[str, np.ndarray], None]:
        return self._unidata.obsm if self._unidata is not None else None

    @obsm.setter
    def obsm(self, obsm: Dict[str, np.ndarray]):
        assert self._unidata is not None
        self._unidata.obsm = obsm

    @property
    def varm(self) -> Union[Dict[str, np.ndarray], None]:
        return self._unidata.varm if self._unidata is not None else None

    @varm.setter
    def varm(self, varm: Dict[str, np.ndarray]):
        assert self._unidata is not None
        self._unidata.varm = varm

    @property
    def uns(self) -> Union[dict, None]:
        return self._unidata.uns if self._unidata is not None else None

    @uns.setter
    def uns(self, uns: dict):
        assert self._unidata is not None
        self._unidata.uns = uns

    def list_keys(self, key_type: str = "matrix") -> List[str]:
        """ Surrogate function for UnimodalData, return available keys in metadata, key_type = barcode, feature, matrix, other
        """
        assert self._unidata is not None
        return self._unidata.list_keys(key_type)

    def select_matrix(self, key: str) -> None:
        """ Surrogate function for UnimodalData, select a matrix
        """
        assert self._unidata is not None
        self._unidata.select_matrix(key)

    def get_matrix(self, key: str) -> csr_matrix:
        """ Surrogate function for UnimodalData, return a matrix indexed by key
        """
        assert self._unidata is not None
        return self._unidata.get_matrix(key)

    def get_exptype(self) -> str:
        """ Surrogate function for UnimodalData, return experiment tpye, can be either 'rna', 'citeseq', 'hashing', 'tcr', 'bcr', 'crispr' or 'atac'.
        """
        assert self._unidata is not None
        return self._unidata.get_exptype()



    def list_data(self) -> List[str]:
        return list(self.data)


    def add_data(self, key: str, uni_data: UnimodalData) -> None:
        """ Add data, if _selected is not set, set as the first added dataset
        """
        if key in self.data:
            raise ValueError("Key {} already exists!".format(key))
        self.data[key] = uni_data
        if self._selected is None:
            self._selected = key
            self._unidata = uni_data


    def select_data(self, key: str) -> None:
        if key not in self.data:
            raise ValueError("Key {} does not exist!".format(key))
        self._selected = key
        self._unidata = self.data[self._selected]


    def current_data(self) -> str:
        return self._selected


    def get_data(self, key: str) -> UnimodalData:
        if key not in self.data:
            raise ValueError("Key {} does not exist!".format(key))
        return self.data[key]


    def drop_data(self, key: str) -> UnimodalData:
        if key not in self.data:
            raise ValueError("Key {} does not exist!".format(key))
        return self.data.pop(key)


    def subset_data(self, data_subset: Set[str] = None, exptype_subset: Set[str] = None) -> None:
        """ Only keep data that are in data_subset and exptype_subset
        """
        if data_subset is not None:
            for key in self.list_data():
                if key not in data_subset:
                    self.data.pop(key)

        if exptype_subset is not None:
            for key in self.list_data():
                if self.data[key].uns["experiment_type"] not in exptype_subset:
                    self.data.pop(key)


    def scan_black_list(self, black_list: Set[str] = None):
        """ Remove unwanted keys in the black list
            Note: black_list might be changed.
        """
        if black_list is None:
            return None

        def _check_reserved_keyword(black_list: Set[str], keyword: str):
            if keyword in black_list:
                logger.warning("Removed reserved keyword '{}' from black list.".format(keyword))
                black_list.remove(keyword)

        _check_reserved_keyword(black_list, "genome")
        _check_reserved_keyword(black_list, "experiment_type")

        for key in self.data:
            self.data[key].scan_black_list(black_list)


    def from_anndata(self, data: anndata.AnnData, genome: str = None, exptype: str = None) -> None:
        """ Initialize from an anndata object
        """
        unidata = UnimodalData(data)
        key = unidata.uns["genome"]
        self.data = {key: unidata}
        self._selected = key
        self._unidata = unidata


    def to_anndata(self) -> anndata.AnnData:
        """ Convert current data to an anndata object
        """
        if self._unidata is None:
            raise ValueError("Please first select a unimodal data to convert!")
        return self._unidata.to_anndata()



# cython: language_level=3

import numpy as np
cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple split_barcode_channel(str[:] arr):
    cdef Py_ssize_t size = arr.size, i

    barcodes = np.empty(size, dtype = np.object)
    channels = np.empty(size, dtype = np.object)

    cdef str[:] bview = barcodes, cview = channels
    cdef list res

    for i in range(size):
        res = arr[i].rsplit(sep = '-', maxsplit = 1)
        bview[i] = res[0]
        if len(res) > 1:
            cview[i] = res[1]

    return (barcodes, channels)



import os
import time
import anndata
from typing import Tuple, Set
import logging
logger = logging.getLogger("pegasusio")

from pegasusio import UnimodalData, MultimodalData

from .hdf5_utils import load_10x_h5_file, load_pegasus_h5_file, load_loom_file, write_loom_file
from .text_utils import load_mtx_file, write_mtx_file, load_csv_file, write_scp_file
from .zarr_utils import ZarrFile



def infer_file_type(input_file: str) -> Tuple[str, str, str]:
    """ Infer file format from input_file name

    This function infer file type by inspecting the file name.

    Parameters
    ----------

    input_file : `str`
        Input file name.

    Returns
    -------
    `str`
        File type, choosing from 'zarr', 'h5sc'(obsoleted), 'h5ad', 'loom', '10x', 'mtx', 'csv' and 'tsv'.
    `str`
        The path covering all input files. Most time this is the same as input_file. But for HCA mtx and csv, this should be parent directory.
    `str`
        Type of the path, either 'file' or 'directory'.

    Note: The last two `str`s are mainly used for transfer to cloud
    """
    file_type = None
    copy_path = input_file
    copy_type = "file"

    if input_file.endswith(".zarr"):
        file_type = "zarr"
    elif input_file.endswith(".h5sc"):
        file_type = "h5sc"
    elif input_file.endswith(".h5ad"):
        file_type = "h5ad"
    elif input_file.endswith(".loom"):
        file_type = "loom"
    elif input_file.endswith(".h5"):
        file_type = "10x"
    elif (
        input_file.endswith(".mtx")
        or input_file.endswith(".mtx.gz")
        or os.path.splitext(input_file)[1] == ""
    ):
        file_type = "mtx"
        if os.path.splitext(input_file)[1] != "":
            copy_path = os.path.dirname(input_file)
        copy_type = "directory"
    elif input_file.endswith(".csv") or input_file.endswith(".csv.gz"):
        file_type = "csv"
        if os.path.basename(input_file) == "expression.csv":
            copy_path = os.path.dirname(input_file)
            copy_type = "directory"
    elif input_file.endswith(".txt") or input_file.endswith(".tsv") or input_file.endswith(
        ".txt.gz") or input_file.endswith(".tsv.gz"):
        file_type = "tsv"
    else:
        raise ValueError("Unrecognized file type for file {}!".format(input_file))

    return file_type, copy_path, copy_type


def read_input(
    input_file: str,
    file_type: str = None,
    genome: str = None,
    exptype: str = None,
    ngene: int = None,
    select_singlets: bool = False,
    black_list: Set[str] = None,
    select_data: Set[str] = None,
    select_exptype: Set[str] = None 
) -> MultimodalData:
    """Load data into memory.

    This function is used to load input data into memory. Inputs can be in 10x genomics v2 & v3 formats (hdf5 or mtx), HCA DCP mtx and csv formats, Drop-seq dge format, and CSV format.

    Parameters
    ----------

    input_file : `str`
        Input file name.
    file_type : `str`, optional (default: None)
        File type, choosing from 'zarr', 'h5sc'(obsoleted), 'h5ad', 'loom', '10x', 'mtx', 'csv', or 'tsv'. If None, inferred from input_file
    genome : `str`, optional (default: None)
        For formats like loom, mtx, dge, csv and tsv, genome is used to provide genome name. In this case if genome is None, except mtx format, "unknown" is used as the genome name instead.
    exptype : `str`, optional (default: None)
        Default experiment type, choosing from 'rna', 'citeseq', 'hashing', 'tcr', 'bcr', 'crispr' or 'atac'. If None, use 'rna' as default.
    ngene : `int`, optional (default: None)
        Minimum number of genes to keep a barcode. Default is to keep all barcodes.
    select_singlets : `bool`, optional (default: False)
        If this option is on, only keep DemuxEM-predicted singlets when loading data.
    black_list : `Set[str]`, optional (default: None)
        Attributes in black list will be poped out.
    select_data: `Set[str]`, optional (default: None)
        Only select data with keys in select_data.
    select_exptype: `Set[str]`, optional (default: None)
        Only select data with experiment type in select_exptype.

    Returns
    -------

    A MultimodalData object.

    Examples
    --------
    >>> data = io.read_input('example_10x.h5')
    >>> data = io.read_input('example.h5ad')
    >>> data = io.read_input('example_ADT.csv', genome = 'hashing_HTO', exptype = 'hashing')
    """
    start = time.perf_counter()

    input_file = os.path.expanduser(os.path.expandvars(input_file))
    if file_type is None:
        file_type, _, _ = infer_file_type(input_file)

    if file_type == "zarr":
        zf = ZarrFile(input_file)
        data = zf.read_multimodal_data(ngene = ngene, select_singlets = select_singlets)
        del zf
    elif file_type == "h5sc":
        data = load_pegasus_h5_file(input_file, ngene=ngene, select_singlets=select_singlets)
    elif file_type == "h5ad":
        data = MultimodalData(anndata.read_h5ad(input_file), genome = genome, exptype = exptype)
    elif file_type == "loom":
        data = load_loom_file(input_file, genome = genome, exptype = exptype, ngene = ngene)
    elif file_type == "10x":
        data = load_10x_h5_file(input_file, ngene=ngene)
    elif file_type == "mtx":
        data = load_mtx_file(input_file, genome = genome, exptype = exptype, ngene = ngene)
    else:
        assert file_type == "csv" or file_type == "tsv"
        data = load_csv_file(input_file, sep = "," if file_type == "csv" else "\t", genome = genome, exptype = exptype, ngene = ngene)

    data.subset_data(select_data, select_exptype)
    data.scan_black_list(black_list)    

    end = time.perf_counter()
    logger.info("{} file '{}' is loaded, time spent = {:.2f}s.".format(file_type, input_file, end - start))

    return data



def write_output(
    data: MultimodalData,
    output_file: str,
    file_type: str = None,
    zarr_zipstore: bool = False,
    is_sparse: bool = True,
    precision: int = 2
) -> None:
    """ Write data back to disk.

    This function is used to write data back to disk.

    Parameters
    ----------

    data : MutimodalData
        data to write back.
    output_file : `str`
        output file name. Note that for mtx files, output_file specifies a directory. For scp format, file_type must be specified. 
    file_type : `str`, optional (default: None)
        File type can be 'zarr', 'h5ad', 'loom', 'mtx' or 'scp'. If file_type is None, it will be inferred based on output_file.
    zarr_zipstore: `bool`, optional (default: False)
        Only apply to zarr output, using ZipStore to have one file instead of a folder.
    is_sparse : `bool`, optional (default: True)
        Only used for writing out SCP-compatible files, if write expression as a sparse matrix.
    precision : `int`, optional (default: 2)
        Precision after decimal point for values in mtx and scp expression matrix.

    Returns
    -------
    `None`

    Examples
    --------
    >>> io.write_output(data, 'test.zarr')
    """
    start = time.perf_counter()

    output_file = os.path.expanduser(os.path.expandvars(output_file))

    def _infer_output_file_type(output_File: str) -> str:
        if output_file.endswith(".zarr"):
            return "zarr"
        elif output_file.endswith(".h5ad"):
            return "h5ad"
        elif output_file.endswith(".loom"):
            return "loom"
        else:
            return "mtx"

    file_type = _infer_output_file_type(output_file) if file_type is None else file_type

    if file_type == "zarr":
        zf = ZarrFile(output_file, mode = "w", storage_type = "ZipStore" if zarr_zipstore else None)
        zf.write_multimodal_data(data)
        del zf
    elif file_type == "h5ad":
        data.to_anndata().write(output_file, compression="gzip")
    elif file_type == "loom":
        write_loom_file(data, output_file)
    elif file_type == "mtx":
        write_mtx_file(data, output_file, precision = precision)
    elif file_type == "scp":
        write_scp_file(data, output_file, is_sparse = is_sparse, precision = precision)
    else:
        raise ValueError("Unknown file type {}!".format(file_type))

    end = time.perf_counter()
    logger.info("{} file '{}' is written. Time spent = {:.2f}s.".format(file_type, output_file, end - start))



import gc
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from collections.abc import MutableMapping
from typing import List, Dict, Union, Set

import logging
logger = logging.getLogger(__name__)

import anndata
from pegasusio.cylib.funcs import split_barcode_channel



class UnimodalData:
    def __init__(
        self,
        barcode_metadata: Union[dict, pd.DataFrame, anndata.AnnData] = None,
        feature_metadata: Union[dict, pd.DataFrame] = None,
        matrices: Dict[str, csr_matrix] = None,
        barcode_multiarrays: Dict[str, np.ndarray] = None,
        feature_multiarrays: Dict[str, np.ndarray] = None,
        metadata: dict = None,
    ) -> None:
        if isinstance(barcode_metadata, anndata.AnnData):
            self.from_anndata(barcode_metadata)
            return None

        def replace_none_df(value):
            return value if value is not None else pd.DataFrame()

        def replace_none(value):
            """ Set empty dictionary as default is dangerous since others might fill in values to the dict
            """
            return value if value is not None else dict()

        self.barcode_metadata = replace_none_df(barcode_metadata) # barcode metadata
        self.feature_metadata = replace_none_df(feature_metadata) # feature metadata
        self.matrices = replace_none(matrices) # a dictionary of scipy csr matrix
        self.barcode_multiarrays = replace_none(barcode_multiarrays)
        self.feature_multiarrays = replace_none(feature_multiarrays)
        self.metadata = replace_none(metadata)  # other metadata, a dictionary
        self.cur_matrix = "X" # default matrix is X

        if len(self.barcode_metadata) > 0:
            if isinstance(self.barcode_metadata, MutableMapping):
                if "barcodekey" not in self.barcode_metadata:
                    raise ValueError("Cannot locate barcode index barcodekey!")
                barcodekey = self.barcode_metadata.pop("barcodekey")
                self.barcode_metadata = pd.DataFrame(data = self.barcode_metadata, index = pd.Index(barcodekey, name = "barcodekey"))
            else:
                if not isinstance(self.barcode_metadata, pd.DataFrame):
                    raise ValueError("Unknown barcode_metadata type: {}!".format(type(self.barcode_metadata)))
                if "barcodekey" in self.barcode_metadata:
                    self.barcode_metadata.set_index("barcodekey", inplace=True)
                elif self.barcode_metadata.index.name != "barcodekey":
                    raise ValueError("Cannot locate barcode index barcodekey!")

        if len(self.feature_metadata) > 0:
            if isinstance(self.feature_metadata, MutableMapping):
                self.feature_metadata = pd.DataFrame(self.feature_metadata)
            if not isinstance(self.feature_metadata, pd.DataFrame):
                raise ValueError("Unknown feature_metadata type: {}!".format(type(self.feature_metadata)))
            if "featurekey" in self.feature_metadata:
                self.polish_featurekey(self.feature_metadata, self.metadata.get('genome', None))
                self.feature_metadata.set_index("featurekey", inplace=True)
            elif self.feature_metadata.index.name != "featurekey":
                raise ValueError("Cannot locate feature index featurekey!")

        if (self.X is not None) and (self.barcode_metadata.shape[0] != self.X.shape[0]):
            raise ValueError(
                "Wrong number of cells : matrix has {} cells, barcodes file has {}".format(
                    self.X.shape[0], self.barcode_metadata.shape[0]
                )
            )
        if (self.X is not None) and (self.feature_metadata.shape[0] != self.X.shape[1]):
            raise ValueError(
                "Wrong number of features : matrix has {} features, features file has {}".format(
                    self.X.shape[1], self.feature_metadata.shape[0]
                )
            )


    def __repr__(self) -> str:
        repr_str = "UnimodalData object with n_obs x n_vars = {} x {}".format(self.barcode_metadata.shape[0], self.feature_metadata.shape[0])
        repr_str += "\n    It contains {} matrices: {}".format(len(self.matrices), str(list(self.matrices))[1:-1])
        repr_str += "\n    It currently binds to matrix '{}' as X\n".format(self.cur_matrix) if len(self.matrices) > 0 else "\n    It currently binds to no matrix\n"
        for key in ["obs", "var", "obsm", "varm", "uns"]:
            repr_str += "\n    {}: {}".format(key, str(list(getattr(self, key).keys()))[1:-1])

        return repr_str


    @property
    def obs(self) -> pd.DataFrame:
        return self.barcode_metadata

    @obs.setter
    def obs(self, obs: pd.DataFrame):
        assert obs.shape[0] == 0 or obs.index.name == "barcodekey"
        self.barcode_metadata = obs

    @property
    def obs_names(self) -> pd.Index:
        return self.barcode_metadata.index

    @obs_names.setter
    def obs_names(self, obs_names: pd.Index):
        self.barcode_metadata.index = obs_names
        self.barcode_metadata.index.name = "barcodekey"

    @property
    def var(self) -> pd.DataFrame:
        return self.feature_metadata

    @var.setter
    def var(self, var: pd.DataFrame):
        assert var.shape[0] == 0 or var.index.name == "featurekey"
        self.feature_metadata = var

    @property
    def var_names(self) -> pd.Index:
        return self.feature_metadata.index

    @var_names.setter
    def var_names(self, var_names: pd.Index):
        self.feature_metadata.index = var_names
        self.feature_metadata.index.name = "featurekey"

    @property
    def X(self) -> Union[csr_matrix, None]:
        return self.matrices.get(self.cur_matrix, None)

    @X.setter
    def X(self, X: csr_matrix):
        self.matrices[self.cur_matrix] = X

    @property
    def obsm(self) -> Dict[str, np.ndarray]:
        return self.barcode_multiarrays

    @obsm.setter
    def obsm(self, obsm: Dict[str, np.ndarray]):
        self.barcode_multiarrays = obsm

    @property
    def varm(self) -> Dict[str, np.ndarray]:
        return self.feature_multiarrays

    @varm.setter
    def varm(self, varm: Dict[str, np.ndarray]):
        self.feature_multiarrays = varm

    @property
    def uns(self) -> dict:
        return self.metadata

    @uns.setter
    def uns(self, uns: dict):
        self.metadata = uns


    def polish_featurekey(self, feature_metadata: pd.DataFrame, genome: str) -> None:
        """    Remove prefixing genome strings and deduplicate feature keys
        """
        if genome is not None:
            # remove genome strings for 10x chemistry if > 1 genomes exist
            import re
            prefix = re.compile("^{}_+".format(genome))
            if prefix.match(feature_metadata["featurekey"][0]):
                feature_metadata["featurekey"] = np.array([prefix.sub("", x) for x in feature_metadata["featurekey"].values], dtype = object)
                if "featureid" in feature_metadata:
                    feature_metadata["featureid"] = np.array([prefix.sub("", x) for x in feature_metadata["featureid"].values], dtype = object)

        if feature_metadata["featurekey"].duplicated().sum() > 0:
            # deduplicate feature keys
            keys = feature_metadata["featurekey"].values
            ids = feature_metadata["featureid"].values if "featureid" in feature_metadata else None

            from collections import Counter
            dup_ids = Counter()

            for i in range(keys.size):
                idn = dup_ids[keys[i]]
                if idn > 0:
                    # key only is duplicated
                    if ids is not None:
                        keys[i] = keys[i] + '_' + ids[i]
                        idn = dup_ids[keys[i]]
                idn += 1                    
                dup_ids[keys[i]] = idn
                if idn > 1:
                    keys[i] = keys[i] + ".#~{}".format(idn) # duplicate ID starts from 2. .#~ makes it unique.


    def get_exptype(self) -> str:
        """ return experiment tpye, can be either 'rna', 'citeseq', 'hashing', 'tcr', 'bcr', 'crispr' or 'atac'.
        """
        return self.metadata.get("experiment_type", None)


    def get_genome(self) -> str:
        """ return genome key
        """
        return self.metadata.get("genome", None)


    def list_keys(self, key_type: str = "matrix") -> List[str]:
        """ Return available keys in metadata, key_type = barcode, feature, matrix, other
        """
        if key_type == "barcode":
            return [
                self.barcode_metadata.index.name
            ] + self.barcode_metadata.columns.tolist()
        elif key_type == "feature":
            return [
                self.feature_metadata.index.name
            ] + self.feature_metadata.columns.tolist()
        elif key_type == "matrix":
            return list(self.matrices)
        elif key_type == "other":
            return list(self.metadata)
        else:
            raise ValueError("Unknown type {}!".format(type))


    def current_matrix(self) -> str:
        """ Return current matrix as self.X
        """
        return self.cur_matrix


    def add_matrix(self, key: str, mat: csr_matrix) -> None:
        """ Add a new matrix, can be raw count or others
        """
        if key in self.matrices:
            raise ValueError("Matrix key {} already exists!".format(key))
        if self.barcode_metadata.shape[0] != mat.shape[0]:
            raise ValueError("Wrong number of cells: matrix has {} cells, which is not match with barcode file ({})!".format(mat.shape[0], self.barcode_metadata.shape[0]))
        if self.feature_metadata.shape[0] != mat.shape[1]:
            raise ValueError("Wrong number of features: matrix has {} features, which is not match with feature file ({})!".format(mat.shape[1], self.feature_metadata.shape[0]))

        self.matrices[key] = mat


    def select_matrix(self, key: str) -> None:
        """ Select a matrix for self.X
        """
        if key not in self.matrices:
            raise ValueError("Matrix key {} does not exist!".format(key))
        self.cur_matrix = key


    def get_matrix(self, key: str) -> csr_matrix:
        """ Return a matrix indexed by key
        """
        if key not in self.matrices:
            raise ValueError("Matrix key {} does not exist!".format(key))
        return self.matrices[key]


    def trim(self, selected: List[bool]) -> None:
        """ Only keep barcodes in selected
        """
        self.barcode_metadata = self.barcode_metadata[selected]
        for key, mat in self.matrices.items():
            self.matrices[key] = mat[selected, :]
        gc.collect()


    def filter(self, ngene: int = None, select_singlets: bool = False) -> None:
        """ Filter out low quality barcodes, only keep barcodes satisfying ngene >= ngene and selecting singlets if select_singlets is True
        """
        if (len(self.matrices) == 0) or ((ngene is None) and (not select_singlets)):
            return None

        self.select_matrix("X")
        selected = np.ones(self.X.shape[0], dtype=bool)
        if ngene is not None:
            selected = selected & (self.X.getnnz(axis=1) >= ngene)
        if select_singlets and ("demux_type" in self.barcode_metadata):
            selected = (
                selected & (self.barcode_metadata["demux_type"] == "singlet").values
            )
            self.barcode_metadata.drop(columns="demux_type", inplace=True)

        self.trim(selected)


    def separate_channels(self) -> None:
        """ Separate channel information from barcodekeys, only used for 10x v2, v3 h5 and mtx.
        """
        if self.barcode_metadata.shape[0] == 0:
            return None # no data

        barcodes, channels = split_barcode_channel(self.barcode_metadata.index.values)

        if channels[0] is None:
            return None # no need to separate channel information and the file should not be generated by cellranger

        if (channels != "1").sum() > 0:
            # we have multiple channels
            self.barcode_metadata["Channel"] = channels
            barcodes = np.array([x + "-" + y for x, y in zip(channels, barcodes)], dtype = object)
            
        self.barcode_metadata.index = pd.Index(barcodes, name="barcodekey")


    def update_barcode_metadata_info(
        self, sample_name: str, row: "pd.Series", attributes: List[str]
    ) -> None:
        """ Update barcodekey, update channel and add attributes
        """
        nsample = self.barcode_metadata.shape[0]
        barcodes = [sample_name + "-" + x for x in self.barcode_metadata.index]
        self.barcode_metadata.index = pd.Index(barcodes, name="barcodekey")
        if "Channel" in self.barcode_metadata:
            self.barcode_metadata["Channel"] = [
                sample_name + "-" + x for x in self.barcode_metadata["Channel"]
            ]
        else:
            self.barcode_metadata["Channel"] = np.repeat(sample_name, nsample)
        if attributes is not None:
            for attr in attributes:
                self.barcode_metadata[attr] = np.repeat(row[attr], nsample)


    def scan_black_list(self, black_list: Set[str]):
        """ Remove (key, value) pairs where key is in black_list
        """
        def _scan_dataframe(df: pd.DataFrame, black_list: Set[str]):
            cols = []
            for key in df.columns:
                if key in black_list:
                    cols.append(key)
            if len(cols) > 0:
                df.drop(columns = cols, inplace = True)

        def _scan_dict(mapping: dict, black_list: Set[str]):
            for key in mapping:
                if key in black_list:
                    mapping.pop(key)

        _scan_dataframe(self.barcode_metadata, black_list)
        _scan_dataframe(self.feature_metadata, black_list)

        _scan_dict(self.matrices, black_list)
        _scan_dict(self.barcode_multiarrays, black_list)
        _scan_dict(self.feature_multiarrays, black_list)
        _scan_dict(self.metadata, black_list)


    def from_anndata(self, data: anndata.AnnData, genome: str = None, exptype: str = None) -> None:
        """ Initialize from an anndata object
            If genome/exptype is not None, set 'genome'/'experiment_type' as genome/exptype
        """
        self.barcode_metadata = data.obs
        self.barcode_metadata.index.name = "barcodekey"
        
        self.feature_metadata = data.var
        self.feature_metadata.index.name = "featurekey"
        if "gene_ids" in self.feature_metadata:
            self.feature_metadata.rename(columns = {"gene_ids": "featureid"}, inplace = True)

        def _to_csr(X):
            return X if isinstance(X, csr_matrix) else csr_matrix(X)

        self.matrices = {"X": _to_csr(data.X)}
        if data.raw is not None:
            self.matrices["raw.X"] = _to_csr(data.raw.X)
        for key, value in data.layers.items():
            self.matrices[key] = _to_csr(value)

        self.barcode_multiarrays = dict(data.obsm)

        self.feature_multiarrays = dict(data.varm)

        self.metadata = dict(data.uns)

        if genome is not None:
            self.metadata["genome"] = genome
        elif "genome" not in self.metadata:
            self.metadata["genome"] = "unknown"
        elif isinstance(self.metadata["genome"], np.ndarray):
            assert self.metadata["genome"].ndim == 1
            self.metadata["genome"] = self.metadata["genome"][0]

        if exptype is not None:
            self.metadata["experiment_type"] = exptype
        elif "experiment_type" not in self.metadata:
            self.metadata["experiment_type"] = "rna"

        self.cur_matrix = "X"

    
    def to_anndata(self) -> anndata.AnnData:
        """ Convert to anndata
        """
        raw = None
        if "raw.X" in self.matrices:
            raw = anndata.AnnData(X = self.matrices["raw.X"])

        layers = {}
        for key, value in self.matrices.items():
            if key != "X" and key != "raw.X":
                layers[key] = value
        
        return anndata.AnnData(X = self.matrices.get("X", None),
            obs = self.barcode_metadata,
            var = self.feature_metadata,
            uns = self.metadata,
            obsm = self.barcode_multiarrays,
            varm = self.feature_multiarrays,
            layers = layers,
            raw = raw)



# cython: language_level=3, c_string_type=str, c_string_encoding=default

import numpy as np

from libc.stdlib cimport malloc, free
from libc.stdio cimport fopen, fclose, getline, FILE, fscanf, sscanf, fprintf, printf
from libc.string cimport strncmp, strlen, strtok

cimport cython


cdef const char* header_real = b"%%MatrixMarket matrix coordinate real general"
cdef const char* header_int = b"%%MatrixMarket matrix coordinate integer general"
cdef const char* metadata = b"%metadata_json: {\"software\": \"pegasusio\"}"


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple read_mtx(char* mtx_file):
    cdef FILE* fi = fopen(mtx_file, "r")
    cdef char* line = NULL
    cdef size_t size
    cdef char isreal = 0

    assert getline(&line, &size, fi) >= 0
    if strncmp(line, header_real, strlen(header_real)) == 0:
        isreal = 1
    elif strncmp(line, header_int, strlen(header_int)) == 0:
        isreal = 0
    else:
        raise ValueError("Cannot recognize header line : " + line)

    while getline(&line, &size, fi) >=0:
        if line[0] != b'%':
            break

    cdef size_t M, N, L, i

    assert line[0] != b'%'
    assert sscanf(line, "%zu %zu %zu", &M, &N, &L) >= 0
    free(line)

    row_ind = np.zeros(L, dtype = np.intc)
    col_ind = np.zeros(L, dtype = np.intc)
    data = np.zeros(L, dtype = np.float32 if isreal else np.intc)

    cdef int[:] rview = row_ind, cview = col_ind
    cdef int[:] dview_int
    cdef float[:] dview_float

    cdef int x, y, value_int
    cdef float value_float

    if isreal:
        dview_float = data
        for i in range(L):
            assert fscanf(fi, "%d %d %f", &x, &y, &value_float) >= 0
            rview[i] = x - 1
            cview[i] = y - 1
            dview_float[i] = value_float
    else:
        dview_int = data
        for i in range(L):
            assert fscanf(fi, "%d %d %d", &x, &y, &value_int) >= 0
            rview[i] = x - 1
            cview[i] = y - 1
            dview_int[i] = value_int

    fclose(fi)

    return row_ind, col_ind, data, (M, N)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void write_mtx(char* mtx_file, object data, int[:] indices, int[:] indptr, int M, int N, int precision = 2):
    """ Input is csr_matrix internal representation, cell by gene; Output will be gene by cell
    """
    cdef FILE* fo = fopen(mtx_file, "w")
    cdef str fmt_str = ""
    cdef char is_real = 0

    cdef int[:] data_int
    cdef float[:] data_float

    if data.dtype.kind == 'f':
        fprintf(fo, "%s\n", header_real)
        fmt_str = "%d %d %.{}f\n".format(precision)
        is_real = 1
        data_float = data
    elif data.dtype.kind == 'i':
        fprintf(fo, "%s\n", header_int)
        fmt_str = "%d %d %d\n"
        is_real = 0
        data_int = data
    else:
        raise ValueError("Detected unknown dtype: {}!".format(data.dtype))

    cdef const char* fmt = fmt_str
    cdef Py_ssize_t data_size = data.size
    cdef Py_ssize_t i, j

    fprintf(fo, "%s\n", metadata)
    fprintf(fo, "%d %d %zd\n", N, M, data_size)

    for i in range(M):
        for j in range(indptr[i], indptr[i + 1]):
            if is_real:
                fprintf(fo, fmt, indices[j] + 1, i + 1, data_float[j])
            else:
                fprintf(fo, fmt, indices[j] + 1, i + 1, data_int[j])

    fclose(fo)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int is_zero(char* pch):
    cdef int i = 0
    while pch[i] != 0:
        if pch[i] >= b'1' and pch[i] <= b'9':
            return 0
        i += 1
    return 1

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple read_csv(char* csv_file, char* delimiters):
    cdef FILE* fi = fopen(csv_file, "r")
    cdef char* line = NULL
    cdef char* pch = NULL
    cdef size_t size

    cdef str row_key
    cdef list colnames = [], rownames = [], row_ind = [], col_ind = [], data_list = []
    cdef int M = 0, N = 0
    cdef size_t L = 0
    cdef Py_ssize_t i

    assert getline(&line, &size, fi) >= 0
    pch = strtok(line, delimiters)
    assert pch != NULL
    row_key = pch

    pch = strtok(NULL, delimiters)
    while pch != NULL:
        colnames.append(pch)
        N += 1
        pch = strtok(NULL, delimiters)

    if N == 0:
        raise ValueError("File {} contains no columns!".format(csv_file))

    colnames[N - 1] = colnames[N - 1].rstrip("\n\r")

    while getline(&line, &size, fi) >=0:
        if line[0] < 32 or line[0] == 127:
            continue
        pch = strtok(line, delimiters)
        assert pch != NULL
        rownames.append(pch)
        for i in range(N):
            pch = strtok(NULL, delimiters)
            assert pch != NULL
            if is_zero(pch) == 0:
                row_ind.append(M)
                col_ind.append(i)
                data_list.append(pch)
                if i == N - 1:
                    data_list[L] = data_list[L].rstrip("\r\n")
                L += 1
        M += 1

    free(line)
    fclose(fi)

    try:
        data = np.array(data_list, dtype = np.intc)
    except ValueError:
        data = np.array(data_list, dtype = np.float32)

    return row_ind, col_ind, data, (M, N), row_key, rownames, colnames


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void write_dense(char* output_file, str[:] barcodes, str[:] features, object data, int[:] indices, int[:] indptr, int M, int N, int precision = 2):
    """ Input must be csr_matrix internal representation, gene by cell (X.T.tocsr()); Output will be gene by cell
    """
    cdef FILE* fo = fopen(output_file, "w")
    cdef str fmt_str = ""
    cdef char is_real = 0

    cdef int[:] data_int
    cdef float[:] data_float

    if data.dtype.kind == 'f':
        fmt_str = "\t%.{}f\n".format(precision)
        is_real = 1
        data_float = data
    elif data.dtype.kind == 'i':
        fmt_str = "\t%d"
        is_real = 0
        data_int = data
    else:
        raise ValueError("Detected unknown dtype: {}!".format(data.dtype))

    cdef const char* fmt = fmt_str
    cdef Py_ssize_t i, j, k, fr

    fprintf(fo, "GENE");
    for i in range(N):
        fprintf(fo, "\t%s", <char*>barcodes[i])
    fprintf(fo, "\n")

    for i in range(M):
        fprintf(fo, "%s", <char*>features[i]);
        fr = 0
        for j in range(indptr[i], indptr[i + 1]):
            for k in range(fr, indices[j]):
                fprintf(fo, "\t0")
            if is_real:
                fprintf(fo, fmt, data_float[j])
            else:
                fprintf(fo, fmt, data_int[j])
            fr = indices[j] + 1
        for k in range(fr, N):
            fprintf(fo, "\t0")
        fprintf(fo, "\n")

    fclose(fo)



# This code is inspired by HCA zarr python codes and AnnData zarr codes

import os
import shutil
import numpy as np
import pandas as pd
from pandas.api.types import is_categorical_dtype, is_string_dtype, is_scalar, is_dict_like
from scipy.sparse import csr_matrix, issparse
import zarr
from zarr import Blosc
from natsort import natsorted
from typing import List, Dict, Tuple, Union

from pegasusio import UnimodalData, MultimodalData



CHUNKSIZE = 1000000
COMPRESSOR = Blosc(cname = 'lz4', clevel = 5)


def calc_chunk(shape: tuple) -> tuple:
    ndim = len(shape)
    chunks = [0] * ndim
    ords = np.argsort(shape)
    chunk_size = CHUNKSIZE * 1.0
    chunk_value = -1
    for i, idx in enumerate(ords):
        if chunk_value < 0:
            if shape[idx] ** (ndim - i) < chunk_size:
                chunks[idx] = shape[idx]
                chunk_size /= chunks[idx]
            else:
                chunk_value = int(np.ceil(chunk_size ** (1.0 / (ndim - i))))
                chunks[idx] = chunk_value
        else:
            chunks[idx] = chunk_value

    return tuple(chunks)



class ZarrFile:
    def __init__(self, path: str, mode: str = 'r', storage_type: str = None) -> None:
        """ Initialize a Zarr file, if mode == 'w', create an empty one, otherwise, load from path
        path : `str`, path for the zarr object.
        storage_type : `str`, currently only support 'ZipStore' and 'NestedDirectoryStore'. If None, use 'NestedDirectoryStore' by default.
        """
        self.store = self.root = None

        if storage_type is None:
            storage_type = 'NestedDirectoryStore'

        if mode == 'w':
            # Create a new zarr file
            if storage_type == 'ZipStore':
                if os.path.isdir(path):
                    shutil.rmtree(path)
                self.store = zarr.ZipStore(path, mode = 'w')
            else:
                if os.path.isfile(path):
                    os.unlink(path)
                zarr.NestedDirectoryStore(path)
            self.root = zarr.group(self.store, overwrite = True)
            self.root.attrs['store'] = storage_type
        else:
            # Load existing zarr file
            try:
                self.root = zarr.open(path, mode = mode)
            except ValueError:
                self.store = zarr.ZipStore(path, mode = mode)
                self.root = zarr.open_group(self.store, mode = mode)


    def __del__(self):
        if self.store is not None and hasattr(self.store, 'close'):
            self.store.close()



    def read_csr(self, group: zarr.Group) -> csr_matrix:
        return csr_matrix((group['data'], group['indices'], group['indptr']), shape = group.attrs['shape'])


    def read_series(self, group: zarr.Group, name: str) -> Union[pd.Categorical, np.recarray]:
        if 'ordered' in group[name].attrs:
            # categorical column
            return pd.Categorical.from_codes(group[name][...], categories = group['_categories/{0}'.format(name)][...], ordered = group[name].attrs['ordered'])
        else:
            return group[name][...]


    def read_dataframe(self, group: zarr.Group) -> Union[pd.DataFrame, np.ndarray]:
        if group.attrs['data_type'] == 'data_frame':
            columns = [col for col in group.array_keys() if col != '_index']
            df = pd.DataFrame(data = {col: self.read_series(group, col) for col in columns},
                index = pd.Index(self.read_series(group, '_index'), name = group.attrs['index_name']),
                columns = columns)
            return df
        else:
            array = np.rec.fromarrays([self.read_series(group, col) for col in group.array_keys()],
                names = group.attrs['columns'])
            return array


    def read_array(self, group: zarr.Group, name: str) -> Union[np.ndarray, np.recarray]:
        if name in group.group_keys():
            return self.read_dataframe(group[name])
        else:
            return group[name][...]


    def read_mapping(self, group: zarr.Group) -> dict:
        res_dict = {}

        if 'scalar' in group.attrs:
            res_dict.update(group.attrs['scalar'])
        
        for key in group.array_keys():
            res_dict[key] = self.read_array(group, key)
        
        for key in group.group_keys():
            sub_group = group[key]
            data_type = sub_group.attrs['data_type']
            value = None
            if data_type == 'data_frame' or data_type == 'record_array':
                value = self.read_dataframe(sub_group)
            elif data_type == 'csr_matrix':
                value = self.read_csr(sub_group)
            else:
                assert data_type == 'dict'
                value = self.read_mapping(sub_group)
            res_dict[key] = value
        
        return res_dict


    def read_unimodal_data(self, group: zarr.Group, ngene: int = None, select_singlets: bool = False) -> UnimodalData:
        """ Read UnimodalData
            ngene: filter barcodes with < ngene
            select_singlets: only select singlets
            The above two option only works if experiment_type == "rna"
        """
        unidata = UnimodalData(barcode_metadata = self.read_dataframe(group['barcode_metadata']),
                            feature_metadata = self.read_dataframe(group['feature_metadata']),
                            matrices = self.read_mapping(group['matrices']),
                            barcode_multiarrays = self.read_mapping(group['barcode_multiarrays']),
                            feature_multiarrays = self.read_mapping(group['feature_multiarrays']),
                            metadata = self.read_mapping(group['metadata']))

        assert "genome" in unidata.metadata
        assert "experiment_type" in unidata.metadata
        
        if unidata.metadata["experiment_type"] == "rna":
            unidata.filter(ngene, select_singlets)

        return unidata


    def read_multimodal_data(self, ngene: int = None, select_singlets: bool = False) -> MultimodalData:
        """ Read MultimodalData
            ngene: filter barcodes with < ngene
            select_singlets: only select singlets
            The above two option only works for dataset with experiment_type == "rna"
        """
        data = MultimodalData()
        need_trim = (ngene is not None) or select_singlets
        selected_barcodes = None

        for key, group in self.root.groups():
            unidata = self.read_unimodal_data(group, ngene, select_singlets)
            if need_trim and unidata.uns.get("experiment_type", "rna") == "rna":
                selected_barcodes = unidata.obs_names if selected_barcodes is None else selected_barcodes.union(unidata.obs_names)
            data.add_data(key, unidata)

        if need_trim:
            for key in data.list_data():
                unidata = data.get_data(key)
                if unidata.uns.get("experiment_type", "rna") != "rna":
                    selected = unidata.obs_names.isin(selected_barcodes)
                    unidata.trim(selected)

        return data



    def write_csr(self, group: zarr.Group, name: str, matrix: csr_matrix) -> None:
        sub_group = group.create_group(name, overwrite = True)
        sub_group.attrs.update(data_type = 'csr_matrix', shape = matrix.shape)
        sub_group.create_dataset('data', data = matrix.data, shape = matrix.data.shape, chunks = calc_chunk(matrix.data.shape), dtype = matrix.data.dtype, compressor = COMPRESSOR, overwrite = True)
        sub_group.create_dataset('indices', data = matrix.indices, shape = matrix.indices.shape, chunks = calc_chunk(matrix.indices.shape), dtype = matrix.indices.dtype, compressor = COMPRESSOR, overwrite = True)
        sub_group.create_dataset('indptr', data = matrix.indptr, shape = matrix.indptr.shape, chunks = calc_chunk(matrix.indptr.shape), dtype = matrix.indptr.dtype, compressor = COMPRESSOR, overwrite = True)


    def write_series(self, group: zarr.Group, name: str, array: np.ndarray, data_type: str) -> None:
        if data_type == 'data_frame':
            if is_string_dtype(array):
                keywords = np.unique(array)
                if keywords.size <= array.size / 2.0: # at least half
                    array = pd.Categorical(array, categories = natsorted(keywords))
            if is_categorical_dtype(array):
                # write category keys
                categories = group.require_group('_categories')
                values = array.categories.values
                if isinstance(values[0], bytes):
                    values = np.array([x.decode() for x in values], dtype = object)
                dtype = str if values.dtype.kind == 'O' else values.dtype
                categories.create_dataset(name, data = values, shape = values.shape, chunks = calc_chunk(values.shape), dtype = dtype, compressor = COMPRESSOR, overwrite = True)
                # write codes
                codes_arr = group.create_dataset(name, data = array.codes, shape = array.codes.shape, chunks = calc_chunk(array.codes.shape), dtype = array.codes.dtype, compressor = COMPRESSOR, overwrite = True)
                codes_arr.attrs['ordered'] = array.ordered
                return None

        dtype = str if array.dtype.kind == 'O' else array.dtype
        group.create_dataset(name, data = array, shape = array.shape, chunks = calc_chunk(array.shape), dtype = dtype, compressor = COMPRESSOR, overwrite = True)


    def write_dataframe(self, group: zarr.Group, name: str, df: Union[pd.DataFrame, np.recarray]) -> None:
        data_type = 'data_frame' if isinstance(df, pd.DataFrame) else 'record_array'

        sub_group = group.create_group(name, overwrite = True) 
        attrs_dict = {'data_type' : data_type}
        cols = list(df.columns if data_type == 'data_frame' else df.dtype.names)
        if data_type == 'data_frame':
            attrs_dict['index_name'] = df.index.name if df.index.name is not None else 'index'
            sub_group.create_group('_categories', overwrite = True) # create a group to store category keys for catigorical columns
            self.write_series(sub_group, '_index', df.index.values, data_type)

        for col in cols:
            self.write_series(sub_group, col, (df[col].values if data_type == 'data_frame' else df[col]), data_type)

        sub_group.attrs.update(**attrs_dict)


    def write_array(self, group: zarr.Group, name: str, array: np.ndarray) -> None:
        if array.dtype.kind == 'V':
            self.write_dataframe(group, name, array)
        else:
            dtype = str if array.dtype.kind == 'O' else array.dtype
            group.create_dataset(name, data = array, shape = array.shape, chunks = calc_chunk(array.shape), dtype = dtype, compressor = COMPRESSOR, overwrite = True)


    def write_mapping(self, group: zarr.Group, name: str, mapping: dict) -> None:
        sub_group = group.create_group(name, overwrite = True)
        scalar_dict = {}

        for key, value in mapping.items():
            if is_scalar(value):
                scalar_dict[key] = value
            elif isinstance(value, np.ndarray):
                self.write_array(sub_group, key, value)
            elif isinstance(value, pd.DataFrame):
                self.write_dataframe(sub_group, key, value)
            elif is_dict_like(value):
                self.write_mapping(sub_group, key, value)
            elif issparse(value):
                assert isinstance(value, csr_matrix)
                self.write_csr(sub_group, key, value)
            else:
                # assume value is either list or tuple, converting it to np.ndarray
                self.write_array(sub_group, key, np.array(value))

        attrs_dict = {'data_type' : 'dict'}
        if len(scalar_dict) > 0:
            attrs_dict['scalar'] = scalar_dict
        sub_group.attrs.update(**attrs_dict)



    def write_unimodal_data(self, group: zarr.Group, name: str, data: UnimodalData) -> None:
        """ Write UnimodalData
        """
        sub_group = group.create_group(name, overwrite = True)
        sub_group.attrs['data_type'] = 'UnimodalData'

        self.write_dataframe(sub_group, 'barcode_metadata', data.barcode_metadata)
        self.write_dataframe(sub_group, 'feature_metadata', data.feature_metadata)
        self.write_mapping(sub_group, 'matrices', data.matrices)
        self.write_mapping(sub_group, 'barcode_multiarrays', data.barcode_multiarrays)
        self.write_mapping(sub_group, 'feature_multiarrays', data.feature_multiarrays)
        self.write_mapping(sub_group, 'metadata', data.metadata)


    def write_multimodal_data(self, data: MultimodalData) -> None:
        """ Write MultimodalData
        """
        for key in data.list_data():
            unidata = data.get_data(key)
            self.write_unimodal_data(self.root, key, unidata)
